""" Generator functions for PDELab local/grid function spaces etc. """

from dune.perftool.generation import (domain,
                                      function_mangler,
                                      generator_factory,
                                      include_file,
                                      preamble,
                                      backend,
                                      )
from dune.perftool.pdelab.restriction import restricted_name

from loopy import CallMangleInfo
from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType

from pymbolic.primitives import Variable

import numpy


class LFSChild(FunctionIdentifier):
    def __init__(self, lfs):
        self.lfs = lfs

    def __getinitargs__(self):
        return (self.lfs,)

    @property
    def name(self):
        return '{}.child'.format(self.lfs)


@function_mangler
def lfs_child_mangler(target, func, dtypes):
    if isinstance(func, LFSChild):
        return CallMangleInfo(func.name, (NumpyType(str),), (NumpyType(numpy.int32),))


@preamble
def define_lfs_bound(lfs, bound):
    return 'auto {} = {}.size();'.format(bound, lfs)


def name_lfs_bound(lfs):
    # LFS might either be given by an UFL element or by a string describing its name
    from ufl import FiniteElementBase
    if isinstance(lfs, FiniteElementBase):
        return name_lfs_bound(name_lfs(lfs))

    bound = '{}_size'.format(lfs)
    define_lfs_bound(lfs, bound)
    return bound


@preamble
def using_indices():
    return "using namespace Dune::Indices;"


@preamble
def define_lfs(name, father, child):
    using_indices()
    return "auto {} = child({}, _{});".format(name, father, child)


def lfs_child(lfs, children, shape=None, symmetry=False):
    from pymbolic.primitives import Call, Product, Sum
    # Old pre-TensorElement implementation kept for comaptibility
    if shape is None:
        indices = (Variable(children[0]),)
    else:
        if symmetry and len(children) == 2:
            # I do not want to think about tensors of rank > 2 right now
            i, j = children
            if i > j:
                j, i = i, j
            i = Variable(i)
            j = Variable(j)
            n = len(children)

            indices = (Sum((Product((n - 1, i)), Product((.5, i, 1 - i)), j)),)
        else:
            # If the index is not an int we need to make a variable out of it.
            #
            # Note: ints occur in the sumfactorisation case with
            # vector valued functions (eg. Stokes)
            if not isinstance(children[0], int):
                children = tuple(Variable(c) for c in children)
            indices = (Sum(tuple(Product((Product(tuple(shape[j] for j in range(i))), child)) for i, child in enumerate(children))),)

    return Call(LFSChild(lfs), indices)


@generator_factory(cache_key_generator=lambda e, r, **kw: (e, r))
def name_leaf_lfs(leaf_element, restriction, val=None):
    """ This function just caches leaf lfs names based on the
    element. The resulting local function spaces are useful only
    for size information. OTOH, they are available with just the
    leaf element available (as seen in basis evaluation).
    """
    # This generator function should be prepoluted by the lfs tree traversal,
    # so val should always be None when we actually want the result
    assert val

    return val


@generator_factory(cache_key_generator=lambda e, r, t, **kw: (e, r, t), context_tags=("kernel",))
def name_lfs(element, restriction, tree_path, prefix=None):
    # Omitting the prefix is only valid upon a second call, which will
    # result in a cache hit.
    assert prefix

    def _name_lfs(prefix, tree_path):
        name = prefix
        if len(tree_path) > 0:
            name = name + '_' + '_'.join(str(i) for i in tree_path)
        return name

    name = _name_lfs(prefix, tree_path)
    if len(tree_path) > 0:
        father = _name_lfs(prefix, tree_path[:-1])
        # If this localfunction space is the child of another one, trigger
        # the extraction preamble. Necessary before going into recursion
        # for having the correct (top-down) order of preambles
        define_lfs(name, father, tree_path[-1])

    # Recurse into the given element to define all other local function spaces!
    from ufl import MixedElement
    from ufl.functionview import select_subelement
    from ufl.classes import FixedIndex
    subel = select_subelement(element, tree_path)
    if isinstance(subel, MixedElement):
        for i in range(subel.num_sub_elements()):
            name_lfs(element, restriction, tree_path + (FixedIndex(i),), prefix=prefix)

    # Cache the name for the subelement
    name_leaf_lfs(subel, restriction, val=name)

    # Now return the prefix!
    return name


@generator_factory(cache_key_generator=lambda e, **kw: e)
def type_gfs(element, basetype=None, index_stack=None):
    # Omitting basetype and index_stack is only valid upon a second call,
    # which will result in a cache hit.
    assert basetype
    assert index_stack is not None

    # Additionally, element is expected to be a ufl finite element
    from ufl import FiniteElementBase
    assert isinstance(element, FiniteElementBase)

    # Recurse into the given element to define all other local function spaces!
    from ufl import MixedElement
    from ufl.classes import FixedIndex
    if isinstance(element, MixedElement):
        for i, subelem in enumerate(element.sub_elements()):
            type_gfs(subelem, basetype=basetype, index_stack=index_stack + (FixedIndex(i),))

    if len(index_stack) == 0:
        return basetype
    else:
        include_file("dune/typetree/childextraction.hh", filetag="operatorfile")
        return 'Dune::TypeTree::Child<{},{}>'.format(basetype, ','.join(str(i) for i in index_stack))


def traverse_lfs_tree(arg):
    from dune.perftool.ufl.modified_terminals import ModifiedArgument
    assert isinstance(arg, ModifiedArgument)

    # First we need to determine the basename as given in the signature of
    # this kernel method!
    lfs_basename = name_argumentspace(arg)
    from dune.perftool.pdelab.localoperator import lop_template_gfs
    gfs_basename = lop_template_gfs(arg)

    # Now start recursively extracting local function spaces and fill the cache with
    # all those values. That way we can later get a correct local function space with
    # just the ufl finite element.
    from ufl.classes import MultiIndex
    name_lfs(arg.argexpr.ufl_element(), arg.restriction, MultiIndex(()), prefix=lfs_basename)
    type_gfs(arg.argexpr.ufl_element(), basetype=gfs_basename, index_stack=())


@generator_factory(item_tags=("iname",), cache_key_generator=lambda e, r, c: (e, c), context_tags=("kernel",))
def _lfs_iname(element, restriction, context):
    lfs = name_leaf_lfs(element, restriction)
    bound = name_lfs_bound(lfs)

    name = "{}_{}_index".format(lfs, context)
    domain(name, bound)

    return name


def lfs_iname(element, restriction, count=None, context=''):
    """ Get the iname to iterate over the local function space given by element

    Arguments:
    ----------
    element: ufl.FiniteElementBase
        The finite element this local function space belongs to
    argcount: int
        Use to realize double nesting in case of jacobians
    context: str
        Some generation methods will require you to duplicate an iname for
        a given purpose, see the 'Loops and dependencies' of the loopy docs:
        https://documen.tician.de/loopy/tutorial.html#loops-and-dependencies
    """
    assert not ((context == '') and (count is None))
    if count is not None:
        if context != '':
            context = "{}_{}".format(count, context)
        else:
            context = str(count)
    return _lfs_iname(element, restriction, context)


def lfs_inames(element, restriction, count=None, context=''):
    return (lfs_iname(element, restriction, count, context),)


class LFSLocalIndex(FunctionIdentifier):
    def __init__(self, lfs):
        self.lfs = lfs

    def __getinitargs__(self):
        return (self.lfs,)

    @property
    def name(self):
        return '{}.localIndex'.format(self.lfs)


@function_mangler
def lfs_localindex_mangler(target, func, dtypes):
    if isinstance(func, LFSLocalIndex):
        return CallMangleInfo(func.name, (NumpyType(numpy.int32),), (NumpyType(numpy.int32),))


def name_testfunctionspace(restriction):
    return restricted_name("lfsv", restriction)


def name_trialfunctionspace(restriction):
    return restricted_name("lfsu", restriction)


def name_argumentspace(ma):
    from ufl.classes import Argument, Coefficient
    if isinstance(ma.argexpr, Argument):
        if ma.argexpr.number() == 0:
            return name_testfunctionspace(ma.restriction)
        if ma.argexpr.number() == 1:
            return name_trialfunctionspace(ma.restriction)
    if isinstance(ma.argexpr, Coefficient):
        # Index 0 is reserved for trialfunction, index 1 is reserved for jacobian apply function
        assert ma.argexpr.count() < 2
        return name_trialfunctionspace(ma.restriction)
    # We should never encounter an argument other than 0 or 1
    assert False


def type_testfunctionspace():
    return "LFSV"


def type_trialfunctionspace():
    return "LFSU"
